from . import QUOTES, COMMON_SEPARATORS, DEFAULT_WAIT, \
    MAX_ERROR_PRINTOUT, MAX_DISPLAY_COLS

# =================================================
#      Automatically Reading Structured Arrays     
# =================================================

# Return the type of a string (try converting to int and float)
def get_type(obj):
    # Special case for handling empty strings (default to base type NoneType).
    if (hasattr(obj, "__len__") and (len(obj) == 0)) or (obj is None):
        return type(None)
    # Detect type by tring to cast and seeing what works.
    try:
        int(obj)
        return int
    except ValueError:
        try:
            float(obj)
            return float
        except ValueError:
            return str

# Remove nested separators inside of quotes in the line
def remove_quoted_content(line):
    cleaned_line = ""
    in_quotes = ""
    for char in line:
        # Toggle 'inside' quotes.
        if (len(in_quotes) == 0) and (char in QUOTES):
            in_quotes = char
        elif (char == in_quotes):
            in_quotes = ""
        elif (len(in_quotes) == 0):
            cleaned_line += char
    return cleaned_line

# Remove nested separators inside of quotes in the line
def split_with_quotes(line, sep):
    split_line = []
    cleaned_line = ""
    in_quotes = ""
    for char in line:
        # Toggle 'inside' quotes.
        if (char == in_quotes):
            in_quotes = ""
            if (char in cleaned_line): cleaned_line += char
        elif (len(in_quotes) == 0) and (char in QUOTES):
            in_quotes = char
            if (len(cleaned_line) > 0): cleaned_line += char
        elif (len(in_quotes) == 0) and (char == sep):
            split_line.append(cleaned_line)
            cleaned_line = ""
        else:
            cleaned_line += char
    # Append the last chunk of the line to the split.
    split_line.append(cleaned_line)
    return split_line

# Pre:  "filename" is the name of an existing file (including path)
# Post: "filename" is scanned for viable separators, those are all
#       characters used the same number of times on each line. If at
#       the end of the file there are still multiple candidate
#       separators, they are sorted by most-occuring first. If one of
#       COMMON_SEPARATORS exist in the set, it is used, otherwise the
#       most occurring separator is used.
def detect_separator(filename="<no_provided_file>", verbose=False, opened_file=None):
    import time
    lines = []
    if not opened_file: f = open(filename)
    else:               f = opened_file
    # Identify the potential separators from the first line
    first_line = remove_quoted_content(f.readline())
    separators = {char:first_line.count(char) for char in set(first_line)}
    # Update the user
    last_update = 0
    should_update = lambda: (verbose and ((time.time() - last_update) > DEFAULT_WAIT))
    raw_data = f.readlines()
    # Cycle the file (narrowing down the list of potential separators
    for i,line in enumerate(raw_data):
        # Skip empty lines.
        if (len(line.strip()) == 0): continue
        # Update the user on progress if appropriate
        if should_update():
            last_update = time.time()
            print(f"\r{(100.0*i/len(raw_data)):0.1f}% complete", flush=True, end="")
        # Clean the line of quoted strings.
        line = remove_quoted_content(line)
        line_chars = set(line.strip())
        # Make sure all tracked separators appear the correct
        # number of times in the line
        for char in list(separators.keys()):
            if (line.count(char) != separators[char]):
                separators.pop( char )
        # WARNING: Do *not* break early, to verify that the
        #          separators are used correctly throughout the
        #          entire file!
    # Close the opened file object
    if not opened_file: f.close()
    # Go back to the beginning of the file
    else:               f.seek(0)
    # Get the separators from the dictionary sorted by the decreasing
    # number of occurrences the character has on each line of the data.
    separators = sorted(separators.keys(), key=lambda k: -separators[k])
    # Check to make sure there is a separator detected
    if len(separators) == 0:
        raise(Exception("ERROR: No consistently used separator in file."))
    if verbose:
        print(" Detected the following possible separators:\n  ", separators)
    # Search through the common separators first, using one of them if possible
    for sep in COMMON_SEPARATORS:
        if sep in separators:
            if verbose: print(" Using '%s' as separator."%sep)
            break
    else:
        # Otherwise, use the most occurring character that could be a separator
        sep = separators[0]
        print("WARNING: Using '%s' as a separator, even though it is unexpected."%(sep))
    # Return the automatically detected separator
    return sep

# Pre:  "filename" is the name of a txt file that has a header
#       "sep" is the separator used in the data file
#       "display" is True if the user wants the detected types printed
# Post: A data.Data class containing all data from the file.
def read_data(filename="<no_provided_file>", sep=None, types=None,
              verbose=False, opened_file=None, header=True, sample=None):
    import time
    from .data import Data
    if (sep is None): sep = detect_separator(filename, verbose, opened_file)
    # Get the opened file object
    if (not opened_file):
        # Open the file to get the data
        f = open(filename)
    else:
        f = opened_file
        file_name = f.name
    # Get the first line (the header, if desired).
    if header: header = [n.strip() for n in split_with_quotes(f.readline().strip(),sep)]
    # Get the first line of data and learn the types of each column
    line = [v.strip() for v in split_with_quotes(f.readline().strip(), sep)]
    # Assign a numeric 
    if not header: header = [f"column {i+1}" for i in range(len(line))]
    # Automatically detect types (if necessary)
    type_digression = False
    if types is None:
        types = list(map(get_type, line))
        type_digression = True
        # Pad the "types" to match the length of the header.
        types += [type(None)] * (len(header) - len(types))
    # If a singular type was provided, expand that out to fill the columns.
    elif (type(types) == type): types = [types] * len(header)
    # Initialize our data holder
    data = Data(names=header, types=types)
    # Clean up the first line of data (like the list of types was cleand).
    for i in range(len(line)):
        if (line[i] == ""): line[i] = None
    # Pad the end of the line with "None" values.
    if (len(line) < len(header)):
        line += [None] * max(0,len(header) - len(line))
    # Add the first line to the data store (if it has any data in it).
    if (any(v is not None for v in line)):
        data.append(line)
    # Get the rest of the raw data from the file and close it
    raw_data = f.readlines()
    if (not opened_file):
        # Close the file
        f.close()
    # WARNING: Provided file object left to be closed by caller
    # Print out a nicely formatted description of the detected types
    if verbose:
        print(f"Column names and types: {len(header)}")
        first_line  = ""
        second_line = ""
        for i in range(len(header)):
            # For large numbers of columns, only show front and end.
            if (len(header) > MAX_DISPLAY_COLS):
                if (i == MAX_DISPLAY_COLS // 2):
                    first_line += "  ..."
                    second_line += "  ..."
                    continue
                # Skip everything but the front and end of the list.
                elif min(i, len(header)-i-1) >= MAX_DISPLAY_COLS // 2:
                    continue
            h, t = header[i], types[i]
            length = max(len(h), len(t.__name__))
            first_line += f"  {h:{length}s}"
            second_line += f"  {t.__name__:{length}s}"
        print(first_line)
        print(second_line)
    # Now process the rest of the data, dynamically overwriting old data
    errors = []
    last_update = 0
    should_update = lambda: (verbose and ((time.time() - last_update) > DEFAULT_WAIT))
    # Define an appropriate iterator based on the usage of this "read".
    if sample is None: iterator = enumerate(raw_data)
    else:
        # Randomly sample lines from the file, (we've already read one line).
        from .utilities import random_range
        iterator = ((i, raw_data[i]) for i in sorted(random_range(len(raw_data), count=sample-1)))
    # Cycle through the elements of "raw_data".
    for i,line in iterator:
        # Update the user on progress if appropriate
        if should_update():
            last_update = time.time()
            print("\r%0.1f%% complete"% (100.0*i/len(raw_data)), flush=True, end="")
        # Clean up the line (in case there are nested separators)
        line = line.strip()
        # Skip empty rows.
        if (len(line) == 0): continue
        # Read the line of data into a list format (replace empty string with None)
        # Remove leading and trailing quotes from strings if necessary
        list_line = [el.strip() for el in split_with_quotes(line, sep)]
        # Pad the end of the line with "None" values.
        if (len(list_line) < len(header)):
            list_line += [None] * max(0,len(header) - len(list_line))
        # Check for any NoneType columns in Data, resolve type for those elements if possible.
        for col_index, curr_type in zip(range(len(list_line)), data.types):
            if (curr_type is type(None)):
                this_col_type = get_type(list_line[col_index])
                if (this_col_type is not type(None)):
                    list_line[col_index] = this_col_type(list_line[col_index])
        # Replace any empty strings with "None" for missing values
        for j in range(len(list_line)):
            if (list_line[j] == ""): list_line[j] = None
        digressions = 0
        # Try and add the line of data
        try:
            data.append(list_line)
        except Data.BadValue:
            if type_digression:
                # Get the new types.
                new_types = list(map(get_type, list_line))
                type_order = {int:0, float:1, str:2}
                new_types = [max(old,new, key=lambda v: type_order[v])
                             for (old,new) in zip(types, new_types)]
                # Update the user if appropriate
                if verbose and (digressions < MAX_ERROR_PRINTOUT):
                    print(f"\nType digression because of line {i+3}.")
                    for c in range(len(types)):
                        if (types[c] != new_types[c]):
                            digressions += 1
                            old_type_name = str(types[c]).split("'")[1]
                            new_type_name = str(new_types[c]).split("'")[1]
                            print(f"  Column {c+1} digressed from '{old_type_name}' to '{new_type_name}' due to: {list_line[c]}")
                        if (digressions == MAX_ERROR_PRINTOUT):
                            print(" Suppressing further type digression notices..")
                            break
                # Retype the existing data to match the new types
                data.retype(new_types)
                # Append line that matches new types
                data.append(list_line)
                types = new_types
            else:
                errors.append(i+3)
                if len(errors) <= MAX_ERROR_PRINTOUT:
                    print(f"\nERROR ON LINE {i+3}:\n  {line}")
                if len(errors) == MAX_ERROR_PRINTOUT:
                    print("Suppressing further error output. See all erroneous lines in later message.")
        except Data.BadElement:
            errors.append(i+3)
            if len(errors) <= MAX_ERROR_PRINTOUT:
                print(f"\nERROR ON LINE {i+3}:\n  {line}")
            if len(errors) == MAX_ERROR_PRINTOUT:
                print("Suppressing further error output. See all erroneous lines in later message.")


    if verbose and (last_update != 0): print("\r100.0% complete.")
    if (len(errors) > 0) and (not verbose):
        print( "WARNING: Encountered potentially erroneous lines while\n"+
              f"         reading '{filename}'.\n"+
               "         Consider setting 'verbose=True' and checking errors.")
    if verbose and (len(errors) > MAX_ERROR_PRINTOUT):
        print(f"ERRONEOUS LINES FROM DATA:\n  {errors}")
    # Return the data
    return data



#!python 

# Any python code that should be executed (as if __main__) during
# setup should go here to prepare the module on a new computer.


# Build the unique library.
def build_unique():
    import os, ctypes, fmodpy
    # Configure the compilation.
    _verbose = True
    # _c_compiler = "cc"  # <- on macOS with Clang, no nested functions allowed.
    _c_compiler = fmodpy.config.f_compiler  # <- works with gfortran, assuming true for others too.
    _flags = "-shared -fPIC -O3"  # <- shared library, position independent, O3 optimization.
    # _flags += " -D_UNIQUE_DEBUG_ENABLED"  # <- enable for testing / debugging.
    _flags_by_safety = [
        "-fopenmp -D_UNIQUE_ALLOW_LOCAL_FUNCTIONS", # <- slower for small cases, good for *very* large data.
        "-D_UNIQUE_SINGLE_THREADED -D_UNIQUE_ALLOW_LOCAL_FUNCTIONS", # <- disables OpenMP parallelism and imports.
        "-D_UNIQUE_SINGLE_THREADED", # ^^ enable if compiler supports local functions for variable width comparison.
        # ^^ safest option, single threaded with no local functions.
    ]
    _lib_dir = os.path.dirname(os.path.abspath(__file__))
    _lib_path = os.path.join(_lib_dir, "unique", "unique.so")
    _src_path = os.path.join(_lib_dir, "unique", "unique.c")
    try:
        _unique = ctypes.cdll.LoadLibrary(_lib_path)
    except:
        # Compile the shared library.
        import sys
        # Trial one to compile and load.
        try:
            cmd = f'{_c_compiler} {_flags} {_flags_by_safety[0]} -o "{_lib_path}" "{_src_path}"'
            if _verbose: print("  ", cmd, flush=True)
            os.system(cmd)
            if _verbose: print(flush=True)
            _unique = ctypes.cdll.LoadLibrary(_lib_path)
        except:
            # Trial two to compile and load.
            try:
                cmd = f'{_c_compiler} {_flags} {_flags_by_safety[1]} -o "{_lib_path}" "{_src_path}"'
                if _verbose: print(" ", cmd, flush=True)
                os.system(cmd)
                if _verbose: print(flush=True)
                _unique = ctypes.cdll.LoadLibrary(_lib_path)
            # Trial three to compile and load.
            except:
                cmd = f'{_c_compiler} {_flags} {_flags_by_safety[2]} -o "{_lib_path}" "{_src_path}"'
                if _verbose: print(" ", cmd, flush=True)
                os.system(cmd)
                if _verbose: print(flush=True)
                _unique = ctypes.cdll.LoadLibrary(_lib_path)
    # Return the compiled module.
    return _unique


# Build the AXY model (compiled Fortran code).
def build_axy():
    import os, fmodpy
    # Define the output directory.
    _dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "approximate", "axy")
    # Import the RANDOM module, it is self contained and simple.
    _random = fmodpy.fimport(
        input_fortran_file = os.path.join(_dir, "axy_random.f90"),
        dependencies = ["pcg32.f90"],
        output_dir = _dir
    )
    # Import the AXY module, it has many dependencies and uses BLAS and LAPACK libraries.
    #   Get the directory for the AXY compiled source code.
    _dependencies = ["pcg32.f90", "axy_profiler.f90", "axy_random.f90", "axy_matrix_operations.f90", "axy_sort_and_select.f90", "axy.f90"]
    _path = os.path.join(_dir, "axy.f90")
    _axy_kwargs = dict(
        input_fortran_file = _path,
        dependencies = _dependencies,
        output_dir = _dir,
        blas = True,
        lapack = True,
        omp = True,
        wrap = True,
        verbose = False,
    )
    #   Try and build using (local) defaults for fmodpy compilation.
    try:
        _axy = fmodpy.fimport(**_axy_kwargs)
        # Try calling the underlying library.
        _axy.axy.new_model_config(
            adn=2, ado=3, ads=4, ans=1, ane=6, ade=7,
            mdn=8, mdo=9, mds=10, mns=1, mne=12, mde=13,
            num_threads=1,
            noe=0, doe=0,  # temporarily disabling output embeddings
        )
    #   When local default fail, try specifying exactly what shared library functions are needed.
    except Exception as exc:
        import logging
        logging.warn(f"Encountered execption with default build of AXY, trying again with an explicit library search.\n  {exc}")
        _axy = fmodpy.fimport( **_axy_kwargs,
            link_blas="", link_lapack="", rebuild=True,
            libraries = [_dir] + fmodpy.config.libraries,
            symbols = [
                ("sgemm", "blas"),
                ("sgels", "lapack"),
                ("omp_get_max_threads", "omp")
            ],
        )

    # Return the AXY and RANDOM modules.
    return _axy, _random


# Build the BallTree code.
def build_balltree():
    # from util.approximate.delaunay import delsparse
    import os, fmodpy
    _dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "approximate", "balltree")
    # Ball tree.
    _dependencies = ["swap.f90", "prune.f90", "fast_select.f90", "fast_sort.f90", "ball_tree.f90"]
    _path = os.path.join(_dir, "ball_tree.f90")
    _balltree = fmodpy.fimport(
        _path, dependencies=_dependencies, output_dir=_dir, omp=True,
        libraries = [_dir] + fmodpy.config.libraries,
        symbols = [("omp_get_max_threads", "omp")],
    )
    # Fast sort.
    _dependencies = ["swap.f90", "fast_sort.f90"]
    _path = os.path.join(_dir, "fast_sort.f90")
    _fast_sort = fmodpy.fimport(_path, dependencies=_dependencies, output_dir=_dir)
    # Fast select.
    _dependencies = ["swap.f90", "fast_select.f90"]
    _path = os.path.join(_dir, "fast_select.f90")
    _fast_select = fmodpy.fimport(_path, dependencies=_dependencies, output_dir=_dir)
    # Prune.
    _path = os.path.join(_dir, "prune.f90")
    _prune = fmodpy.fimport(_path, output_dir=_dir)
    return _balltree.ball_tree, _fast_sort.fast_sort, _fast_select.fast_select, _prune.prune


# Build the Delaunay code.
def build_delaunay():
    # from util.approximate.delaunay import delsparse
    import os, fmodpy
    _dependencies = ["real_precision.f90", "blas.f", "lapack.f", "slatec.f", "delsparse.f90"]
    _dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "approximate", "delaunay")
    _path = os.path.join(_dir, "delsparse.f90")
    _delsparse = fmodpy.fimport(
        _path, dependencies=_dependencies, output_dir=_dir,
        omp=True, f_compiler_args="-std=legacy -fPIC -shared -O3",
    )
    return _delsparse
    

# Build the regex library.
def build_regex():
    import os, ctypes
    # Configure for the compilation for the C code.
    _c_compiler = "cc"
    _clib_source = os.path.join(os.path.dirname(os.path.abspath(__file__)), "regex", "regex.c")
    _clib_bin = os.path.join(os.path.dirname(os.path.abspath(__file__)), "regex", "libregex.so")
    try:
        # Import existing compiled library.
        _regex = ctypes.CDLL(_clib_bin)
    except:
        # Compile and import.
        _compile_command = f"{_c_compiler} -O3 -fPIC -shared -o '{_clib_bin}' '{_clib_source}'"
        os.system(_compile_command)
        _regex = ctypes.CDLL(_clib_bin)
    # Return the compiled module.
    return _regex


# Function to run during module setup.
def build_all():
    # Try compiling the 'fmath' library.
    import sys
    _first = sys.path.pop(0)
    try:
        build_unique()
        build_axy()
        build_balltree()
        build_delaunay()
        build_regex()
    except Exception as exc:
        print(f"WARNING: 'tlux' encountered exception while trying to build compiled libraries.")
        print()
        print(exc)
    finally:
        sys.path.insert(0, _first)


# Function for cleaning (before or after a build).
def clean_all():
    import os
    import platform
    import shutil
    import tempfile
    # Get the starting directory.
    start_dir = os.getcwd()
    parent_dir = os.path.dirname(os.path.dirname(__file__))
    os.chdir(parent_dir)
    # Move files from the provided list of paths to a temporary directory and print the directory path.
    paths = [
        # Unique.
        "tlux/unique/unique.so",
        # AXY.
        "tlux/approximate/axy/axy",
        "tlux/approximate/axy/axy_random",
        "tlux/approximate/axy/axy.mod",
        "tlux/approximate/axy/matrix_operations.mod",
        "tlux/approximate/axy/pcg32_module.mod",
        "tlux/approximate/axy/profiler.mod",
        "tlux/approximate/axy/random.mod",
        "tlux/approximate/axy/sort_and_select.mod",
        "tlux/approximate/axy/test/test_axy_random",
        "tlux/approximate/axy/test/test_axy_module",
        "tlux/approximate/axy/test/test_axy_profiler",
        "tlux/approximate/axy/test/test_axy_random_OLD",
        "tlux/approximate/axy/test/test_axy_module_OLD",
        "tlux/approximate/axy/test/test_axy_profiler_OLD",
        # Ball Tree.
        "tlux/approximate/balltree/fast_select",
        "tlux/approximate/balltree/fast_sort",
        "tlux/approximate/balltree/prune",
        "tlux/approximate/balltree/ball_tree",
        # # Delaunay.
        # f"tlux/approximate/delaunay/delsparse",
        # Regex.
        "tlux/regex/libregex.so",
    ]
    # Create a temporary directory
    temp_dir = tempfile.mkdtemp()
    # For each path in the provided list
    for path in paths:
        try:
            # Check if the path exists
            if os.path.exists(path):
                # Maintain the relative path structure within the temporary directory and move the file or directory
                relative_path = os.path.relpath(path)
                destination_path = os.path.join(temp_dir, relative_path)
                os.makedirs(os.path.dirname(destination_path), exist_ok=True)
                print(f"  (tlux.setup) cleaning '{path}' to '{destination_path}'..", flush=True)
                shutil.move(path, destination_path)
        except Exception as e:
            # Handle any exceptions that may occur during the file/directory moving process
            print(f"ERROR(tlux.setup): Exception while moving '{path}': {str(e)}", flush=True)
    # Move back to the original directory.
    os.chdir(start_dir)


# Function for orchestrating setup.
def setup():
    clean_all()
    build_all()




import multiprocessing
import queue
import threading
import time
import traceback

# Class for timing operations. Initialize to start, call to check, use
# the `start` and `stop` methods / attributes to set / observe values.
class Timer:
    #-----------------------------------------------------------------
    #                            Private
    _a = 0
    _b = None
    # Offer a function for explicitly beginning a timer.
    def _begin(self):
        self._a = time.time()
        self._b = None
        return self._a
    # End function for this timer.
    def _end(self):
        self._b = time.time()
        return self._b
    # Return the elapsed time since start.
    def _check(self):
        if   (self._a is None): return 0
        elif (self._b is None): return time.time() - self._a
        else:                   return self._b - self._a
    # Overwrite the "__call__" method of the provided object.
    def _callset(obj, func):
        class Float(type(obj)):
            def __call__(self): return func()
            def __repr__(self): return str(float(self))
        return Float(obj)
    #-----------------------------------------------------------------
    #                             Public
    # 
    # Initialization.
    def __init__(self): self._begin()
    # If not stopped, return elapsed time, otherwise return total time.
    def __call__(self, precision=2): return float(f"{self._check():.{precision-1}e}")
    def __str__(self): return str(self())
    # Return "start time" as attribute, "begin timer" as function.
    @property
    def start(self): return Timer._callset(self._a, self._begin)
    # Return "stop time" as attribute, "end timer, return total" as function.
    @property
    def stop(self):
        if (self._b == None): return Timer._callset(self._end(), self.total)
        else:                 return Timer._callset(self._b,     self.total)
    # Return the "total time from start" if running, return "total time" if finished.
    @property
    def total(self):
        if (self._b == None): return Timer._callset(self._check(), self._check)
        else:                 return Timer._callset(self._b - self._a, lambda: self._b - self._a)


# A decorator class for object types that make them behave asynchronously by
# housing instances in their own process. Returned values for all internal attributes
# and methods are 'queue.Queue' objects that have a '.get' method to explicitly
# retrieve results (or wait if the results are not ready yet).
#
# Example:
# 
#   async_obj_type = Asynchronous(SlowObject) # Create an asynchronous object.
#   obj = async_obj_type(*init_args, **init_kwargs)
#   async_result = obj.slow_method(*method_args, **method_kwargs) # Call a slow or expensive method.
#   ... # Do other operations without waiting for results.
#   result = async_result.get() # Retrieve the result from the object method call.
#
class Asynchronous:
    # The following are for the specification of this decorator type.
    _type: type = type(None)
    _stop_word: str = "__exit__"
    _verbose: bool = True
    # The following are populated for instantiated asynchronous objects of this type.
    _instance: object = None # Local instance for the asychronous jobs.
    _tasks: multiprocessing.Queue = None # Queue holding tasks.
    _results: multiprocessing.Queue = None # Queue holding results.
    _result_ids: dict = None # Dictionary holding identities of returned result queues.
    _type_tasks: multiprocessing.Queue = None # Queue holding type check tasks.
    _type_results: multiprocessing.Queue = None # Queue for type check results.
    _communicator: threading.Thread = None # Thread that communicates in shared memory space.
    _worker: multiprocessing.Process = None # Process that runs asynchronously.
    _type_checker: threading.Thread = None # Thread for type checking (in the asynchronous instance).
    _attributes_callable: dict = None # Dictionary holding attribute names and whether they're callable or not.

    # Initialization of a new Asynchronous type.
    def __init__(self, object_type=type(None), stop_word="__exit__", verbose=True):
        super().__setattr__("_type", object_type)
        super().__setattr__("_stop_word", stop_word)
        super().__setattr__("_verbose", verbose)

    # Wrap "Asynchronous" around the original type name.
    def __repr__(self): return f"<Asynchronous {self.__getattribute__('_type').__name__} instance at {id(self)}>"

    # Create an Asynchronous instance of the target object.
    def __call__(self, *args, **kwargs):
        worker = self.__getattribute__("_worker")
        # Presume this call is an initialization of an Asynchronous object.
        if (worker is None):
            obj = type(self)(
                object_type=self.__getattribute__("_type"),
                stop_word=self.__getattribute__("_stop_word"),
                verbose=self.__getattribute__("_verbose"),
            )
            obj.__getattribute__("_start")(*args, **kwargs)
            return obj
        # Presume this "__call__" was meant for the underlying object.
        else: return self.__getattr__("__call__")

    # Make sure all processes and threads are exited.
    def __del__(self):
        # Terminate the type checker thread (owned by the worker process).
        type_tasks = self.__getattribute__("_type_tasks")
        if (type_tasks is not None):
            type_tasks.put(None)
        # Terminate the worker.
        worker = self.__getattribute__("_worker")
        if ((worker is not None) and (worker.is_alive())):
            tasks = self.__getattribute__("_tasks")
            task.put((None, self.__getattribute__("_stop_word"), (), {}))
            worker.terminate()
        # Terminate the communicator thread (owned by the current process).
        communicator = self.__getattribute__("_communicator")
        if ((communicator is not None) and (communicator.is_alive())):
            results = self.__getattribute__("_results")
            results.put((None, None))
            communicator.join()

    # Set up an Asynchronous instance with appropriate processes and communication channels.
    def _start(self, *args, **kwargs):
        # These local dictionaries will hold the identifiers of the returned Queue
        # objects so they can be populated with results, and a cache dictating which
        # attributes of the asynchronous object instance are callable.
        super().__setattr__("_result_ids", {})
        super().__setattr__("_attributes_callable", {})
        # The following queues handle tasks, results, type tasks, and type results.
        super().__setattr__("_tasks", multiprocessing.Queue())
        super().__setattr__("_results", multiprocessing.Queue())
        super().__setattr__("_type_tasks", multiprocessing.Queue())
        super().__setattr__("_type_results", multiprocessing.Queue())
        # Create an asychronous worker that is independent (in memory and CPU).
        super().__setattr__("_worker", multiprocessing.Process(
            target=self.__getattribute__("_work"),
            daemon=True,
        ))
        self.__getattribute__("_worker").start()
        self.__getattribute__("_tasks").put((args, kwargs))
        # Create a thread (shared memory) that collects multiprocessed results and enqueues them.
        super().__setattr__("_communicator", threading.Thread(
            target=self.__getattribute__("_communicate"),
            daemon=True,
        ))
        communicator = self.__getattribute__("_communicator")
        communicator.start()
        # Get all of the attributes and their callability.
        attributes_callable = self.__getattribute__("_attributes_callable")
        type_results = self.__getattribute__("_type_results")
        attr_name, is_callable = type_results.get()
        while (attr_name is not None):
            attributes_callable[attr_name] = is_callable
            attr_name, is_callable = type_results.get()

    # This method is meant to be executed by a thread in the asynchronous process,
    # it can be used to check the type of the attributes of the instance.
    def _check_type(self):
        type_tasks = self.__getattribute__("_type_tasks")
        type_results = self.__getattribute__("_type_results")
        instance = self.__getattribute__("_instance")
        verbose = self.__getattribute__("_verbose")
        # On initialization, put all of the known attributes and their callability.
        for attr_name in dir(instance):
            type_results.put((attr_name, callable(getattr(instance, attr_name))))
        type_results.put((None, None))
        # Go into the listening loop, in case another attribute's callability status is needed.
        while True:
            attr_name = type_tasks.get()
            if (attr_name is None): break
            try: attr_type = callable(getattr(instance, attr_name))
            except Exception as exception: attr_type = exception
            type_results.put(attr_type)

    # Communicate the results of the asynchronous calls back into the local queues.
    # This method shares memory space with the parent process, so it can asynchronously
    # put results into the local 'queue.Queue' objects that were already returned.
    def _communicate(self):
        results = self.__getattribute__("_results")
        result_ids = self.__getattribute__("_result_ids")
        while True:
            identity, result = results.get()
            # When the identity is None, that means the communicator should exit.
            if identity is None:
                break
            # We have a result, put it into the appropriate Queue.
            else:
                result_ids.pop(identity).put(result)

    # Work on the assigned tasks from the parent process. This method will run
    # in its own process isolated from the parent by memory and CPU time.
    def _work(self):
        # Get the task queue.
        tasks = self.__getattribute__("_tasks")
        # Create a local instance of the (now asynchronous) type.
        args, kwargs = tasks.get()
        super().__setattr__("_instance", self.__getattribute__("_type")(*args, **kwargs))
        # Start the type checker thread.
        super().__setattr__("_type_checker", threading.Thread(
            target=self.__getattribute__("_check_type"),
            daemon=True,
        ))
        self._type_checker.start()
        # Get other local attributes that will be reused while working.
        results = self.__getattribute__("_results")
        instance = self.__getattribute__("_instance")
        verbose = self.__getattribute__("_verbose")
        while True:
            identity, name, args, kwargs = tasks.get()
            if identity is None: break
            # Get the attribute and do the appropriate action.
            try:
                attr = getattr(instance, name)
                # For callable attributes, execute the method.
                if callable(attr): results.put((identity, attr(*args, **kwargs)))
                # For value-only attributes, just return the value.
                else: results.put((identity, attr))
            # When an exception is raised, put it into the queue of results.
            except Exception as exception:
                if verbose: traceback.print_exc()
                results.put((identity, exception))
        # Terminate the type checker.
        type_checker = self.__getattribute__("_type_checker")
        type_checker.terminate()

    # This method captures all attempts to access an internal attribute of the instance
    # and redirects the request to the isolated process that the instance is running
    # inside of. If the specified attribute is callable, an asynchronous method is returned
    # that genereates 'queue.Queue' objects as results to calls. If the attribute is not
    # callable, then a 'queue.Queue' object is returned that will contain the value.
    def __getattr__(self, name):
        tasks = self.__getattribute__("_tasks")
        # Check to see if the instance attribute is callable.
        attributes_callable = self.__getattribute__("_attributes_callable")
        is_callable = attributes_callable.get(name, None)
        # If we do not already know the callability of the attribute, make a request to check.
        if (is_callable is None):
            type_tasks = self.__getattribute__("_type_tasks")
            type_results = self.__getattribute__("_type_results")
            type_tasks.put(name)
            is_callable = type_results.get()
            if isinstance(is_callable, Exception): raise(is_callable)
            else: attributes_callable[name] = is_callable
        # Fetch the queue that will hold the identities of the results.
        result_ids = self.__getattribute__("_result_ids")
        # If the attribute is callable, generate an asynchronous wrapper function and return it.
        if is_callable:
            def asynchronous_method(*args, **kwargs):
                async_result = queue.Queue()
                result_ids[id(async_result)] = async_result
                tasks.put((id(async_result), name, args, kwargs))
                return async_result
            return asynchronous_method
        # Else, this attribute just has a value and that value needs to be returned as a Queue.
        else:
            async_result = queue.Queue()
            result_ids[id(async_result)] = async_result
            tasks.put((id(async_result), name, tuple(), dict()))
            return async_result

    # Set an attribute on the asynchronous instance.
    def __setattr__(self, name, value):
        raise(NotImplementedError)



import os
import numpy as np
from tlux.approximate.base import Approximator


# Wrapper class for using the Delaunay fortran code
class Delaunay(Approximator):
    def __init__(self, *args, **kwargs):
        # Set up the algorithm for parallel or serial evaluation.
        try:
            # # TODO: Enable nested parallization.
            # os.environ["OMP_NESTED"] = "TRUE"
            # from tlux.approximate.delaunay.delsparse import delaunaysparsep as delaunaysparse
            from tlux.approximate.delaunay.delsparse import delaunaysparses as delaunaysparse
        except:
            from tlux.setup import build_delaunay
            delsparse = build_delaunay()
            delaunaysparse = delsparse.delaunaysparses
        self.delaunay = delaunaysparse
        # Initialize containers.
        self.x = None
        self.y = None
        self.errs = {}
        # If this model was initialized with data, assume it was a fit.
        if (max(len(args),len(kwargs)) > 0):
            self.fit(*args, **kwargs)

    # Use fortran code to compute the boxes for the given data
    def _fit(self, x, y=None):
        self.x = np.asarray(x, dtype="float64", order="C").T
        if (y is not None):
            self.y = np.asarray(y, dtype="float64", order="C").T
        self.errs = {}

    # Return just the points and the weights associated with them for
    # creating the correct interpolation
    def _predict(self, x, allow_extrapolation=True, print_errors=True,
                 interpolate=True, eps=2**(-26), ibudget=10000,
                 extrap=100.0, **kwargs):
        # Get the predictions from Delaunay Fortran code.
        d = self.x.shape[0]
        n = self.x.shape[1]
        pts = np.array(self.x, order="F")
        m = x.shape[0]
        q = np.array(x.T, dtype="float64", order="F")
        simps = np.ones(shape=(d+1, m), dtype="int32", order="F")
        weights = np.ones(shape=(d+1, m), dtype="float64", order="F")
        ierr = np.ones(shape=(m,), dtype="int32", order="F")
        if ((self.y is not None) and (interpolate)):
            ir = self.y.shape[0]
            interp_in = np.array(self.y, order="F")
            y = np.zeros(shape=(ir,m), dtype="float64", order="F")
        else: interp_in = y = None
        self.delaunay(d, n, pts, m, q, simps, weights, ierr,
                      interp_in=interp_in, interp_out=y,
                      ibudget=ibudget, eps=eps, extrap=extrap,
                      **kwargs)
        # Check for extrapolation.
        assert (allow_extrapolation or ((ierr==1).sum() == 0)), "Encountered extrapolation points when making Delaunay prediction."
        ierr = np.where(ierr == 1, 0, ierr)
        # Handle any errors that may have occurred.
        if (ierr.sum() > 0):
            if print_errors:
                unique_errors = sorted(np.unique(ierr))
                print(" [Delaunay errors:",end="")
                for e in unique_errors:
                    if (e == 0): continue
                    self.errs[e] = np.arange(len(ierr))[ierr == e]
                    print(" %3i"%e,"at","{"+",".join(tuple(
                        str(i) for i in range(len(ierr))
                        if (ierr[i] == e)))+"}", end=";")
                print("] ")
            # Reset the errors to simplex of 1s (to be 0) and weights of 0s.
            bad_indices = (ierr > 0)
            if ((self.y is not None) and (interpolate)):
                y[:,bad_indices] = 0.0
            else:
                simps[:,bad_indices] = 0
                weights[:,bad_indices] = 0.0
        # Return the appropriate values.
        if ((self.y is not None) and (interpolate)):
            # Return the values.
            return y.T
        else:
            # Return the weights and indices.
            return simps-1, weights

    # Get the weights and indices of the source points from the Delaunay
    #  simplicial mesh that recreate the provided points 'x'.
    def weights(self, *args, **kwargs):
        kwargs["interpolate"] = False
        return self.predict(*args, **kwargs)


# Define a simple test to make sure it is working.
if __name__ == "__main__":
    from tlux.plot import Plot
    from tlux.random import well_spaced_box, well_spaced_ball

    # Generate some data.
    x = well_spaced_box(400, 2)
    y = np.cos(10*np.linalg.norm(x, axis=1))

    # Fit a nearest neighbor model.
    m = Delaunay()
    # from tlux.approximate.axy import AXY
    # m = AXY()
    m.fit(x=x, y=y)

    # Create a visual.
    p = Plot()
    p.add("data", *x.T, y)
    p.add_func("fit", m, *([[0,1]]*x.shape[1]), vectorized=True, plot_points=5000)
    p.show(z_range=[-2,4])



import fmodpy
import os
import numpy as np
_this_dir = os.path.dirname(os.path.abspath(__file__))
# Build a class that contains pointers to the model internals, allowing
#  python attribute access to all of the different components of the models.
class AposModel:
    def __init__(self, config, model):
        self.config = config
        self.model = model
        self.a_embeddings  = self.model[self.config.asev-1:self.config.aeev].reshape(self.config.ade, self.config.ane, order="F")
        self.a_input_vecs  = self.model[self.config.asiv-1:self.config.aeiv].reshape(self.config.adi, self.config.ads, order="F")
        self.a_input_shift = self.model[self.config.asis-1:self.config.aeis].reshape(self.config.ads, order="F")
        self.a_state_vecs  = self.model[self.config.assv-1:self.config.aesv].reshape(self.config.ads, self.config.ads, max(0,self.config.ans-1), order="F")
        self.a_state_shift = self.model[self.config.asss-1:self.config.aess].reshape(self.config.ads, max(0,self.config.ans-1), order="F")
        self.a_output_vecs = self.model[self.config.asov-1:self.config.aeov].reshape(self.config.adso, self.config.ado, order="F")
        self.m_embeddings  = self.model[self.config.msev-1:self.config.meev].reshape(self.config.mde, self.config.mne, order="F")
        self.m_input_vecs  = self.model[self.config.msiv-1:self.config.meiv].reshape(self.config.mdi, self.config.mds, order="F")
        self.m_input_shift = self.model[self.config.msis-1:self.config.meis].reshape(self.config.mds, order="F")
        self.m_state_vecs  = self.model[self.config.mssv-1:self.config.mesv].reshape(self.config.mds, self.config.mds, max(0,self.config.mns-1), order="F")
        self.m_state_shift = self.model[self.config.msss-1:self.config.mess].reshape(self.config.mds, max(0,self.config.mns-1), order="F")
        self.m_output_vecs = self.model[self.config.msov-1:self.config.meov].reshape(self.config.mdso, self.config.mdo, order="F")
        self.ax_shift = self.model[self.config.aiss-1:self.config.aise]
        self.ay_shift = self.model[self.config.aoss-1:self.config.aose]
        self.x_shift = self.model[self.config.miss-1:self.config.mise]
        self.y_shift = self.model[self.config.moss-1:self.config.mose]
    # Allow square brackets to access attributes of this model and its configuration.
    def __getitem__(self, attr):
        if hasattr(self, attr):
            return getattr(self, attr)
        elif hasattr(self.config, attr):
            return getattr(self.config, attr)
    # Create a summary string for this model.
    def __str__(self, vecs=False):
        # A function for creating a byte-size string from an integer.
        def _byte_str(byte_size):
            if (byte_size < 2**10):
                byte_size = f"{byte_size} bytes"
            elif (byte_size < 2**20):
                byte_size = f"{byte_size//2**10:.1f}KB"
            elif (byte_size < 2**30):
                byte_size = f"{byte_size//2**20:.1f}MB"
            elif (byte_size < 2**40):
                byte_size = f"{byte_size//2**30:.1f}GB"
            else:
                byte_size = f"{byte_size//2**40:.1f}TB"
            return byte_size
        # Calculate the byte size of this model (excluding python descriptors).
        # TODO: Not all configs are 4 bytes, do more expensive sum over actual sizes?
        byte_size = len(self.config._fields_)*4 + self.model.dtype.itemsize*self.model.size
        byte_size = _byte_str(byte_size)
        if (self.config.rwork_size+self.config.iwork_size > 0):
            work_size = self.config.rwork_size*4 + self.config.iwork_size*4
            byte_size += " + "+_byte_str(work_size)+" work space"
        # Create a function that prints the actual contents of the arrays.
        if vecs: to_str = lambda arr: "\n    " + "\n    ".join(str(arr).split("\n")) + "\n"
        else:    to_str = lambda arr: "\n"
        # Provide details (and some values where possible).
        return (
            f"APOS model ({self.config.total_size} parameters) [{byte_size}]\n"+
            (" apositional\n"+
            f"  input dimension  {self.config.adn}\n"+
            f"  output dimension {self.config.ado}\n"+
            f"  state dimension  {self.config.ads}\n"+
            f"  number of states {self.config.ans}\n"+
           (f"  embedding dimension  {self.config.ade}\n"+
            f"  number of embeddings {self.config.ane}\n"
             if self.config.ane > 0 else "")+
            f"  embeddings   {self.a_embeddings.shape}  "+to_str(self.a_embeddings)+
            f"  input vecs   {self.a_input_vecs.shape}  "+to_str(self.a_input_vecs)+
            f"  input shift  {self.a_input_shift.shape} "+to_str(self.a_input_shift)+
            f"  state vecs   {self.a_state_vecs.shape}  "+to_str(self.a_state_vecs)+
            f"  state shift  {self.a_state_shift.shape} "+to_str(self.a_state_shift)+
            f"  output vecs  {self.a_output_vecs.shape} "+to_str(self.a_output_vecs)+
             "\n" if (self.a_output_vecs.size > 0) else "") +
            (" positional\n"+
            f"  input dimension  {self.config.mdn}\n"+
            f"  output dimension {self.config.mdo}\n"+
            f"  state dimension  {self.config.mds}\n"+
            f"  number of states {self.config.mns}\n"+
           (f"  embedding dimension  {self.config.mde}\n"+
            f"  number of embeddings {self.config.mne}\n"
             if self.config.mne > 0 else "")+
            f"  embeddings   {self.m_embeddings.shape}  "+to_str(self.m_embeddings)+
            f"  input vecs   {self.m_input_vecs.shape}  "+to_str(self.m_input_vecs)+
            f"  input shift  {self.m_input_shift.shape} "+to_str(self.m_input_shift)+
            f"  state vecs   {self.m_state_vecs.shape}  "+to_str(self.m_state_vecs)+
            f"  state shift  {self.m_state_shift.shape} "+to_str(self.m_state_shift)+
            f"  output vecs  {self.m_output_vecs.shape} "+to_str(self.m_output_vecs)
             if (self.m_output_vecs.size > 0) else "")
        )
# Class for calling the underlying APOS model code.
class APOS:
    # Make the string function return the unpacked model.
    def __str__(self): return str(self.unpack())
    # Initialize a new APOS model.
    def __init__(self, source="apos.f90", name=None, **kwargs):
        try:
            import fmodpy
            apos = fmodpy.fimport(source, name=name, blas=True,
                                  lapack=True, omp=True, wrap=True,
                                  verbose=False, output_dir=_this_dir,
            )
            # Store the Fortran module as an attribute.
            self.APOS = apos.apos
        except:
            # TODO:
            #  - python fallback that supports the basic evaluation of
            #    a model (but no support for training new models).
            raise(NotImplementedError("The Fortran source was not loaded successfully."))
        # Set defaults for standard internal parameters.
        self.steps = 1000
        self.seed = None
        self.num_threads = None
        self.config = None
        self.model = np.zeros(0, dtype="float32")
        self.record = np.zeros(0, dtype="float32")
        # Default descriptors for categorical inputs.
        self.axi_map = []
        self.axi_sizes = []
        self.axi_starts = []
        self.xi_map = []
        self.xi_sizes = []
        self.xi_starts = []
        self.yi_map = []
        self.yi_sizes = []
        self.yi_starts = []
        # Initialize the attributes of the model that can be initialized.
        self._init_model(**kwargs)
    # Initialize a model, if possible.
    def _init_model(self, **kwargs):
        # Apositional model parameters.
        adn = kwargs.pop("adn", 0)
        ado = kwargs.pop("ado", None)
        ads = kwargs.pop("ads", None)
        ans = kwargs.pop("ans", None)
        ade = kwargs.pop("ade", None)
        ane = kwargs.pop("ane", None)
        # Model parameters.
        mdn = kwargs.pop("mdn", None)
        mdo = kwargs.pop("mdo", None)
        mds = kwargs.pop("mds", None)
        mns = kwargs.pop("mns", None)
        mde = kwargs.pop("mde", None)
        mne = kwargs.pop("mne", None)
        # Number of threads.
        self.num_threads = kwargs.pop("num_threads", self.num_threads)
        self.seed = kwargs.pop("seed", self.seed)
        self.steps = kwargs.pop("steps", self.steps)
        # Initialize if enough arguments were provided.
        if (None not in {adn, mdn, mdo}):
            self.config = self.APOS.new_model_config(
                adn=adn, ado=ado, ads=ads, ans=ans, ane=ane, ade=ade,
                mdn=mdn, mdo=mdo, mds=mds, mns=mns, mne=mne, mde=mde,
                num_threads=self.num_threads)
            # Set any configuration keyword arguments given at initialization
            #  that were not passed to "new_model_config".
            for n in ({n for (n,t) in self.config._fields_} & set(kwargs)):
                setattr(self.config, n, kwargs[n])
            # Set all internal arrays and initialize the model.
            self.model = np.zeros(self.config.total_size, dtype="float32")
            self.APOS.init_model(self.config, self.model, seed=self.seed)
    # Generate the string containing all the configuration information for this model.
    def config_str(self):
        s = ""
        max_n_len = max(map(len,(n for (n,t) in self.config._fields_)))
        max_t_len = max(map(len,(str(t).split("'")[1].split('.')[1]
                                 for (n,t) in self.config._fields_)))
        for (n,t) in self.config._fields_:
            t = str(t).split("'")[1].split('.')[1]
            s += f"  {str(t):{max_t_len}s}  {n:{max_n_len}s}  =  {getattr(self.config,n)}\n"
        return s
    # Unpack the model (which is in one array) into it's constituent parts.
    def unpack(self):
        # If there is no model or configuration, return None.
        if (self.config is None) or (self.model is None):
            return None
        return AposModel(self.config, self.model)
    # Given a categorical input array, construct a dictionary for
    #  mapping the unique values in the columns of the array to integers.
    def _i_map(self, xi):
        if (len(xi.dtype) > 0):
            xi_map = [np.unique(xi[n]) for n in xi.dtype.names]
        else:
            xi_map = [np.unique(xi[:,i]) for i in range(xi.shape[1])]
        xi_sizes = [len(u) for u in xi_map]
        xi_starts = (np.cumsum(xi_sizes) - xi_sizes[0] + 1).tolist()
        return xi_map, xi_sizes, xi_starts
    # Given a categorical input array (either 2D or struct), map this
    #  array to an integer encoding matrix with the same number of
    #  columns, but unique integers assigned to each unique value.
    def _i_encode(self, xi, xi_map, xi_sizes, xi_starts):
        xi_rows = xi.shape[0]
        xi_cols = len(xi.dtype) or xi.shape[1]
        _xi = np.zeros((xi_rows, xi_cols), dtype="int32", order="C")
        for i in range(xi_cols):
            start_index = xi_starts[i]
            num_unique = xi_sizes[i]
            unique_vals = xi_map[i]
            vals = (xi[:,i:i+1] if len(xi.dtype) == 0 else xi[xi.dtype.names[i]])
            eq_val = vals == unique_vals
            # Add a column to the front that is the default if none match.
            eq_val = np.concatenate((
                np.logical_not(eq_val.max(axis=1)).reshape(xi_rows,1),
                eq_val), axis=1)
            val_indices = np.ones((xi_rows,num_unique+1), dtype="int32") * np.arange(num_unique+1)
            val_indices[:,1:] += start_index-1
            _xi[:,i] = val_indices[eq_val]
        return _xi
    # Convert all inputs to the APOS model into the expected numpy format.
    def _to_array(self, y, yi, x, xi, ax, axi, sizes):
        # Get the number of inputs.
        if   (y  is not None): nm = len(y)
        elif (yi is not None): nm = len(yi)
        elif (x  is not None): nm = len(x)
        elif (xi is not None): nm = len(xi)
        elif (sizes is not None): nm = len(sizes)
        # Make sure that all inputs are numpy arrays.
        if (y is not None):  y = np.asarray(y, dtype="float32", order="C")
        else:                y = np.zeros((nm,0), dtype="float32", order="C") 
        if (yi is not None): yi = np.asarray(yi)
        else:                yi = np.zeros((nm,0), dtype="int32", order="C")
        if (x is not None): x = np.asarray(x, dtype="float32", order="C")
        else:               x = np.zeros((nm,0), dtype="float32", order="C")
        if (xi is not None): xi = np.asarray(xi)
        else:                xi = np.zeros((nm,0), dtype="int32", order="C")
        if (sizes is not None): sizes = np.asarray(sizes, dtype="int32")
        else:                   sizes = np.zeros(0, dtype="int32")
        na = sizes.sum()
        if (ax is not None): ax = np.asarray(ax, dtype="float32", order="C")
        else:                ax = np.zeros((na,0), dtype="float32", order="C")
        if (axi is not None): axi = np.asarray(axi)
        else:                 axi = np.zeros((na,0), dtype="int32", order="C")
        # Make sure that all inputs have the expected shape.
        assert (len(y.shape) in {1,2}), f"Bad y shape {y.shape}, should be 1D or 2D matrix."
        assert (len(yi.shape) in {1,2}), f"Bad yi shape {yi.shape}, should be 1D or 2D matrix."
        assert (len(x.shape) in {1,2}), f"Bad x shape {x.shape}, should be 1D or 2D matrix."
        assert (len(xi.shape) in {1,2}), f"Bad xi shape {xi.shape}, should be 1D or 2D matrix."
        assert (len(ax.shape) in {1,2}), f"Bad ax shape {ax.shape}, should be 1D or 2D matrix."
        assert (len(axi.shape) in {1,2}), f"Bad axi shape {axi.shape}, should be 1D or 2D matrix."
        assert (len(sizes.shape) == 1), f"Bad sizes shape {sizes.shape}, should be 1D int vectora."
        # Reshape inputs to all be two dimensional (except sizes).
        if (len(y.shape) == 1): y = y.reshape((-1,1))
        if (len(yi.shape) == 1) and (len(yi.dtype) == 0): yi = yi.reshape((-1,1))
        if (len(x.shape) == 1): x = x.reshape((-1,1))
        if (len(xi.shape) == 1) and (len(xi.dtype) == 0): xi = xi.reshape((-1,1))
        if (len(ax.shape) == 1): ax = ax.reshape((-1,1))
        if ((len(axi.shape) == 1) and (len(axi.dtype) == 0)): axi = axi.reshape((-1,1))
        mdo = y.shape[1]
        mdn = x.shape[1]
        adn = ax.shape[1]
        # Handle mapping "xi" into integer encodings.
        xi_cols = len(xi.dtype) or xi.shape[1]
        if (xi_cols > 0):
            if (len(self.xi_map) == 0):
                self.xi_map, self.xi_sizes, self.xi_starts = self._i_map(xi)
            else:
                assert (xi_cols == len(self.xi_map)), f"Bad number of columns in 'xi', {xi_cols}, expected {len(self.xi_map)} columns."
            xi = self._i_encode(xi, self.xi_map, self.xi_sizes, self.xi_starts)
            mne = sum(self.xi_sizes)
        else: mne = 0
        # Handle mapping "axi" into integer encodings.
        axi_cols = len(axi.dtype) or axi.shape[1]
        if (axi_cols > 0):
            if (len(self.axi_map) == 0):
                self.axi_map, self.axi_sizes, self.axi_starts = self._i_map(axi)
            else:
                assert (axi_cols == len(self.axi_map)), f"Bad number of columns in 'axi', {axi_cols}, expected {len(self.axi_map)} columns."
            axi = self._i_encode(axi, self.axi_map, self.axi_sizes, self.axi_starts)
            ane = sum(self.axi_sizes)
        else: ane = 0
        # Handle mapping "yi" into integer encodings.
        yi_cols = len(yi.dtype) or yi.shape[1]
        if (yi_cols > 0):
            if (len(self.yi_map) == 0):
                self.yi_map, self.yi_sizes, self.yi_starts = self._i_map(yi)
            else:
                assert (yi_cols == len(self.yi_map)), f"Bad number of columns in 'yi', {yi_cols}, expected {len(self.yi_map)} columns."
            yi = self._i_encode(yi, self.yi_map, self.yi_sizes, self.yi_starts)
            yne = sum(self.yi_sizes)
        else: yne = 0
        # Handle mapping integer encoded "yi" into a single real valued y.
        if (yne > 0):
            embedded = np.concatenate((
                np.zeros((1,yne), dtype="float32"),
                np.identity(yne, dtype="float32")), axis=0)
            _y = np.zeros((nm, mdo+yne), dtype="float32")
            _y[:,:mdo] = y[:,:]
            for i in range(yi.shape[1]):
                _y[:,mdo:] += embedded[yi[:,i]]
            y = _y
            mdo += yne
        # Return all the shapes and numpy formatted inputs.
        return nm, na, mdn, mne, mdo, adn, ane, yne, y, x, xi, ax, axi, sizes
    # Fit this model.
    def fit(self, x=None, y=None, yi=None, xi=None, ax=None, axi=None,
            sizes=None, new_model=False, **kwargs):
        # Ensure that 'y' values were provided.
        assert ((y is not None) or (yi is not None)), "APOS.fit requires 'y' or 'yi' values, but neitherwere provided (use keyword argument 'y=<values>' or 'yi=<values>')."
        # Make sure that 'sizes' were provided for apositional (aggregate) inputs.
        if ((ax is not None) or (axi is not None)):
            assert (sizes is not None), "APOS.fit requires 'sizes' to be provided for apositional input sets (ax and axi)."
        # Get all inputs as arrays.
        nm, na, mdn, mne, mdo, adn, ane, yne, y, x, xi, ax, axi, sizes = (
            self._to_array(y, yi, x, xi, ax, axi, sizes)
        )
        # Configure this model if requested (or not already done).
        if (new_model or (self.config is None)):
            # Ensure that the config is compatible with the data.
            kwargs.update({
                "adn":adn,
                "ane":max(ane, kwargs.get("ane",0)),
                "mdn":mdn,
                "mne":max(mne, kwargs.get("mne",0)),
                "mdo":mdo,
            })
            if (max(kwargs["mdn"], kwargs["mne"]) == 0):
                kwargs["ado"] = kwargs["mdo"]
                kwargs["mdo"] = 0
                kwargs["mns"] = 0
            self._init_model(**kwargs)
        # If there are integer embeddings, expand "x" and "ax" to have space to hold those embeddings.
        if (self.config.ade > 0):
            _ax = np.zeros((ax.shape[0],ax.shape[1]+self.config.ade), dtype="float32", order="C")
            _ax[:,:ax.shape[1]] = ax
            ax, _ax = _ax, ax
        if (self.config.mde > 0) or (self.config.ado > 0):
            _x = np.zeros((x.shape[0],self.config.mdi), dtype="float32", order="C")
            _x[:,:x.shape[1]] = x
            x, _x = _x, x
        # ------------------------------------------------------------
        # If a random seed is provided, then only 2 threads can be used
        #  because nondeterministic behavior comes from reordered addition.
        if (self.seed is not None):
            if (self.config.num_threads > 2):
                import warnings
                warnings.warn("Seeding an APOS model will deterministically initialize weights, but num_threads > 2 will result in a nondeterministic model fit.")
        # Get the number of steps for training.
        steps = kwargs.get("steps", self.steps)
        # ------------------------------------------------------------
        # Set up new work space for this minimization process.
        self.APOS.new_fit_config(nm, na, self.config)
        self.rwork = np.zeros(self.config.rwork_size, dtype="float32")
        self.iwork = np.zeros(self.config.iwork_size, dtype="int32")
        # Minimize the mean squared error.
        self.record = np.zeros((steps,6), dtype="float32", order="C")
        try:
            result = self.APOS.minimize_mse(self.config, self.model, self.rwork, self.iwork,
                                            ax.T, axi.T, sizes, x.T, xi.T, y.T, 
                                            steps=steps, record=self.record.T)
        except:
            yw = np.zeros((nm,0), dtype="float32", order="C")
            result = self.APOS.minimize_mse(self.config, self.model, self.rwork, self.iwork,
                                            ax.T, axi.T, sizes, x.T, xi.T, y.T, yw.T,
                                            steps=steps, record=self.record.T)
        assert (result[-1] == 0), f"APOS.minimize_mse returned nonzero exit code {result[-1]}."
        # Copy the updated values back into the input arrays (for transparency).
        if (self.config.mde > 0):
            _x[:,:] = x[:,:_x.shape[1]]
        if (self.config.ade > 0):
            _ax[:,:] = ax[:,:_ax.shape[1]]
    # Calling this model is an alias for 'APOS.predict'.
    def __call__(self, *args, **kwargs):
        return self.predict(*args, **kwargs)
    # Make predictions for new data.
    def predict(self, x=None, xi=None, ax=None, axi=None, sizes=None,
                embedding=False, save_states=False, **kwargs):
        # Evaluate the model at all data.
        assert ((x is not None) or (xi is not None) or (sizes is not None)), "APOS.predict requires at least one of 'x', 'xi', or 'sizes' to not be None."
        # Make sure that 'sizes' were provided for apositional (aggregate) inputs.
        if ((ax is not None) or (axi is not None)):
            assert (sizes is not None), "APOS.predict requires 'sizes' to be provided for apositional input sets (ax and axi)."
        # Make sure that all inputs are numpy arrays.
        nm, na, mdn, mne, mdo, adn, ane, yne, _, x, xi, ax, axi, sizes = (
            self._to_array(None, None, x, xi, ax, axi, sizes)
        )
        # Embed the inputs into the purely positional form.
        ade = self.config.ade
        ads = self.config.ads
        ado = self.config.ado
        mde = self.config.mde
        mds = self.config.mds
        if (self.config.mdo != 0):
            mdo = self.config.mdo
        else:
            mdo = self.config.ado
        # Compute the true real-vector input dimensions given embeddings.
        adn += ade
        mdn += mde + ado
        # ------------------------------------------------------------
        # Initialize storage for all arrays needed at evaluation time.
        #   If there are integer embeddings, expand "ax" and "x" to have
        #   space to hold those embeddings.
        if (self.config.ade > 0):
            _ax = np.zeros((ax.shape[0],ax.shape[1]+self.config.ade), dtype="float32", order="C")
            _ax[:,:ax.shape[1]] = ax
            ax = _ax
        ay = np.zeros((na, ado), dtype="float32", order="F")
        if (self.config.mde > 0) or (self.config.ado > 0):
            _x = np.zeros((x.shape[0],self.config.mdi), dtype="float32", order="C")
            _x[:,:x.shape[1]] = x
            x = _x
        y = np.zeros((nm, mdo), dtype="float32", order="C")
        if (save_states):
            m_states = np.zeros((nm, mds, self.config.mns), dtype="float32", order="F")
            a_states = np.zeros((na, ads, self.config.ans), dtype="float32", order="F")
        else:
            m_states = np.zeros((nm, mds, 2), dtype="float32", order="F")
            a_states = np.zeros((na, ads, 2), dtype="float32", order="F")
        # ------------------------------------------------------------
        # Call the unerlying library.
        info = self.APOS.check_shape(self.config, self.model, ax.T, axi.T, sizes, x.T, xi.T, y.T)
        assert (info == 0), f"APOS.predict encountered nonzero exit code {info} when calling APOS.check_shape."
        self.APOS.embed(self.config, self.model, axi.T, xi.T, ax.T, x.T)
        result = self.APOS.evaluate(self.config, self.model, ax.T, ay, sizes,
                                    x.T, y.T, a_states, m_states, info)
        assert (result[-1] == 0), f"APOS.evaluate returned nonzero exit code {result[-1]}."
        # Save the states if that's 
        if (save_states):
            self.a_states = a_states
            self.ay = ay
            self.m_states = m_states
        # If there are embedded y values in the output, return them to the format at training time.
        if (len(self.yi_map) > 0) and (not embedding):
            yne = sum(self.yi_sizes)
            _y = [y[:,i] for i in range(y.shape[1]-yne)]
            for i in range(len(self.yi_map)):
                start = self.yi_starts[i]
                size = self.yi_sizes[i]
                _y.append(
                    self.yi_map[i][np.argmax(y[:,start:start+size], axis=1)]
                )
            return np.asarray(_y).T
        elif (embedding and (len(self.yi_map) == 0)):
            return m_states[:,:,-2]
        else:
            return y
    # Save this model to a path.
    def save(self, path):
        import json
        with open(path, "w") as f:
            # Get the config as a Python type.
            if (self.config is None): config = None
            else: config = {n:getattr(self.config, n) for (n,t) in self.config._fields_}
            # Write the JSON file with Python types.
            f.write(json.dumps({
                # Create a dictionary of the known python attributes.
                "config" : config,
                "model"  : self.model.tolist(),
                "record" : self.record.tolist(),
                "xi_map"    : [l.tolist() for l in self.xi_map],
                "xi_sizes"  : self.xi_sizes,
                "xi_starts" : self.xi_starts,
                "axi_map"    : [l.tolist() for l in self.axi_map],
                "axi_sizes"  : self.axi_sizes,
                "axi_starts" : self.axi_starts,
                "yi_map"    : [l.tolist() for l in self.yi_map],
                "yi_sizes"  : self.yi_sizes,
                "yi_starts" : self.yi_starts,
            }))
    # Load this model from a path (after having been saved).
    def load(self, path):
        # Read the file.
        import json
        with open(path, "r") as f:
            attrs = json.loads(f.read())
        # Load the attributes of the model.
        for key in attrs:
            value = attrs[key]
            if (key[-4:] == "_map"):
                value = [np.asarray(l) for l in value]
            elif (key[:2] in {"xi_","axi_","yi_"}):
                pass
            elif (type(value) is list): 
                value = np.asarray(value, dtype="float32")
            setattr(self, key, value)
        # Convert the the dictionary configuration into the correct type.
        if (type(self.config) is dict):
            self.config = self.APOS.MODEL_CONFIG(**self.config)
        # Return self in case an assignment was made.
        return self





import fmodpy
fmodpy.config.f_compiler_args = "-fPIC -shared -O3 -fcheck=bounds"
np.random.seed(0)

# m = 10
# kwargs = dict(
#     adn = 3*m,
#     ane = 8*m,
#     ans = 2*m,
#     ads = 2*m,
#     ado = 3*m,
#     mdn = 5*m,
#     mne = 24*m,
#     mns = 2*m,
#     mds = 3*m,
#     mdo = 2*m,
#     seed = 0,
#     num_threads = 1,
# )
# aold = APOS(source="apos_0-0-15.f90", name="apos_old", **kwargs)
# anew = APOS(source="apos_0-0-19.f90", name="apos_new", **kwargs)
# an = 200
# mn = 100
# nec = 2 # num embedding columns
# sizes = np.ones(mn, dtype=np.int32) * max(1,an // mn)
# sizes[-1] = an - sum(sizes[:-1])
# ax = np.random.random(size=(an, aold.config.adn)).astype(np.float32)
# axi = np.random.randint(1, aold.config.ane // nec, size=(an,nec))
# x = np.random.random(size=(mn, aold.config.mdn)).astype(np.float32)
# xi = np.random.randint(1, aold.config.mne // nec, size=(mn,nec))
# y = np.random.random(size=(mn, aold.config.mdo)).astype(np.float32)


aold = APOS(source="apos_0-0-15.f90", name="apos_old", seed=0, num_threads=1)
anew = APOS(source="apos_0-0-20.f90", name="apos_new", seed=0, num_threads=1)

from tlux.random import well_spaced_box
n = 100
# A function for testing approximation algorithms.
def f(x):
    x = x.reshape((-1,2))
    x, y = x[:,0], x[:,1]
    return (3*x + np.cos(8*x)/2 + np.sin(5*y))
x = well_spaced_box(n, 2)
y = f(x)
y = np.concatenate((y, np.cos(np.linalg.norm(x,axis=1))), axis=0).reshape(-1,1)
# Create all data.
xi = np.concatenate((np.ones(len(x)),2*np.ones(len(x)))).reshape((-1,1)).astype("int32")
ax = np.concatenate((x, x), axis=0).reshape((-1,1)).copy()
axi = (np.ones(ax.shape, dtype="int32").reshape((xi.shape[0],-1)) * (np.arange(xi.shape[1])+1)).reshape(-1,1)
sizes = np.ones(xi.shape[0], dtype="int32") * 2
x = np.zeros((xi.shape[0], 0), dtype="float32", order="C")
# Initialize the models.
aold.fit(ax=ax.copy(), axi=axi, sizes=sizes, x=x.copy(), xi=xi, y=y.copy(), steps=0, num_threads=1)
anew.fit(ax=ax.copy(), axi=axi, sizes=sizes, x=x.copy(), xi=xi, y=y.copy(), steps=0, num_threads=1)


print()
print(aold)
# print(aold.unpack().__str__(vecs=True))
print()
print(anew)
# print(anew.unpack().__str__(vecs=True))
print()

print("aold.config.num_vars: ", aold.config.num_vars)
print("anew.config.num_vars: ", anew.config.num_vars)
nv = max(aold.config.num_vars, anew.config.num_vars)
print("model_difference:", max(abs(aold.model[:nv] - anew.model[:nv])))

# Make the models the same before taking a fit step.
aold.model[:nv] = anew.model[:nv]


# Set internal parameters to be similar (to control code execution path).
# aold.config.ax_normalized = False
# anew.config.ax_normalized = False
# aold.config.ay_normalized = False
# anew.config.ay_normalized = False
# aold.config.x_normalized = False
# anew.config.x_normalized = False
# aold.config.y_normalized = False
# anew.config.y_normalized = False
# aold.config.logging_step_frequency = 1
# anew.config.logging_step_frequency = 1 

# anew.config.orthogonalizing_step_frequency = 100
anew.config.basis_replacement = False
# anew.config.equalize_y = False

# aold.config.encode_normalization = True
# anew.config.encode_normalization = True
aold.config.num_threads = 1
anew.config.num_threads = 1

steps = 100

ax_old = ax.copy()
x_old = x.copy()
y_old = y.copy()
aold.fit(ax=ax_old, axi=axi, sizes=sizes, x=x_old, xi=xi, y=y_old, steps=steps)

ax_new = ax.copy()
x_new = x.copy()
y_new = y.copy()
anew.fit(ax=ax_new, axi=axi, sizes=sizes, x=x_new, xi=xi, y=y_new, steps=steps)
# 
print("model_difference:", max(abs(aold.model[:nv] - anew.model[:nv])))
print()
fy_old = aold(ax=ax_old.copy(), axi=axi, sizes=sizes, x=x_old.copy(), xi=xi, save_states=True)
fy_new = anew(ax=ax_new.copy(), axi=axi, sizes=sizes, x=x_new.copy(), xi=xi, save_states=True)
print("ax_difference: ", abs(ax_old - ax_new).max())
print("as_difference: ", abs(aold.a_states - anew.a_states).max())
print("ay_difference: ", abs(aold.ay - anew.ay).max())
if (x_old.size > 0):
    print("x_difference:  ", abs(x_old - x_new).max())
print("ms_difference: ", abs(aold.m_states - anew.m_states).max())
print("fy_difference: ", abs(fy_old - fy_new).max())
print()

print("ax_shift:   ", abs(aold.rwork[aold.config.aiss-1:aold.config.aise] - 
                          anew.rwork[anew.config.aiss-1:anew.config.aise]).max())
print("ay_shift:   ", abs(aold.rwork[aold.config.aoss-1:aold.config.aose] - 
                          anew.rwork[anew.config.aoss-1:anew.config.aose]).max())
print("ay:         ", abs(aold.rwork[aold.config.say-1:aold.config.eay] - 
                          anew.rwork[anew.config.say-1:anew.config.eay]).max())
print("y_rescale:  ", abs(aold.rwork[aold.config.syr-1:aold.config.eyr] - 
                          anew.rwork[anew.config.syr-1:anew.config.eyr]).max())
print("y_gradient: ", abs(aold.rwork[aold.config.syg-1:aold.config.eyg] -
                          anew.rwork[anew.config.syg-1:anew.config.eyg]).max())
print()

if (min(aold.config.ade, anew.config.ade) > 0):
    print("a_embeddings:  ", abs(aold.unpack().a_embeddings  - anew.unpack().a_embeddings).max() )
print("a_input_vecs:  ", abs(aold.unpack().a_input_vecs  - anew.unpack().a_input_vecs).max() )
print("a_input_shift: ", abs(aold.unpack().a_input_shift - anew.unpack().a_input_shift).max())
print("a_state_vecs:  ", abs(aold.unpack().a_state_vecs  - anew.unpack().a_state_vecs).max() )
print("a_state_shift: ", abs(aold.unpack().a_state_shift - anew.unpack().a_state_shift).max())
print("a_output_vecs: ", abs(aold.unpack().a_output_vecs - anew.unpack().a_output_vecs).max())
print("m_embeddings:  ", abs(aold.unpack().m_embeddings  - anew.unpack().m_embeddings).max() )
print("m_input_vecs:  ", abs(aold.unpack().m_input_vecs  - anew.unpack().m_input_vecs).max() )
print("m_input_shift: ", abs(aold.unpack().m_input_shift - anew.unpack().m_input_shift).max())
print("m_state_vecs:  ", abs(aold.unpack().m_state_vecs  - anew.unpack().m_state_vecs).max() )
print("m_state_shift: ", abs(aold.unpack().m_state_shift - anew.unpack().m_state_shift).max())
print("m_output_vecs: ", abs(aold.unpack().m_output_vecs - anew.unpack().m_output_vecs).max())
print()



import os

# ----------------------------------------------------------------
#  Enable debugging option "-fcheck=bounds".
import fmodpy
# fmodpy.configure(verbose=True)
fmodpy.config.f_compiler_args = "-fPIC -shared -O3 -pedantic -fcheck=bounds -ftrapv -ffpe-trap=invalid,overflow,underflow,zero"
# fmodpy.config.link_blas = "-framework Accelerate"
# fmodpy.config.link_lapack = "-framework Accelerate"
_dependencies = ["random.f90", "matrix_operations.f90", "sort_and_select.f90", "axy.f90"]
_dir = os.path.dirname(os.path.realpath(__file__))
_path = os.path.join(_dir, "axy.f90")
_axy = fmodpy.fimport(
    _path, dependencies=_dependencies, output_dir=_dir,
    blas=True, lapack=True, omp=True, wrap=True,
    rebuild=False,
    # verbose=True, 
    # link_blas="", link_lapack="",
    # libraries = [_dir] + fmodpy.config.libraries,
    # symbols = [
    #     ("sgemm", "blas"),
    #     ("sgels", "lapack"),
    #     ("omp_get_max_threads", "omp")
    # ],
# )
# ----------------------------------------------------------------
